package com.inspur.fabric.config;

import com.inspur.fabric.core.entity.*;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.bouncycastle.asn1.pkcs.PrivateKeyInfo;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.openssl.PEMParser;
import org.bouncycastle.openssl.jcajce.JcaPEMKeyConverter;
import org.hyperledger.fabric.sdk.helper.Utils;
import org.hyperledger.fabric_ca.sdk.HFCAClient;
import org.hyperledger.fabric_ca.sdk.exception.EnrollmentException;
import org.hyperledger.fabric_ca.sdk.exception.InvalidArgumentException;

import java.io.*;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PrivateKey;
import java.security.spec.InvalidKeySpecException;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static java.lang.String.format;

/**
 * @author zhanglan
 * @description
 * @date 2021/8/13
 */
@Slf4j
public class PropertiesConfigService implements IConfigService {

    // fabric证书私钥等文件的根目录
//	public static final String CHANNEL_PATH = "/channel";

//	private static final String DEFAULT_CONFIG = "/fabric.properties";

    private BaasProperties baasProperties;

    private String ordererDomain;

    private static final String PROPBASE = "org.hyperledger.fabric.sdk.";

    private static final String INTEGRATIONTESTS_ORG = PROPBASE + "integration.org.";

    private static final Pattern orgPat = Pattern.compile("^" + Pattern.quote(INTEGRATIONTESTS_ORG) + "([^\\.]+)\\.mspid$");

    private static final Properties sdkProperties = new Properties();

    private static final Map<String, OrgEntity> orgs = new ConcurrentHashMap<String, OrgEntity>();

    private static final Map<String, PeerEntity> peers = new ConcurrentHashMap<String, PeerEntity>();

    private static final Map<String, OrdererEntity> orderers = new ConcurrentHashMap<String, OrdererEntity>();

    public PropertiesConfigService(BaasProperties baasProperties) {
        this.baasProperties = baasProperties;
        try {
//			ClassPathResource resource = new ClassPathResource("fabric.properties");
            File file = new File(baasProperties.getConfigFilePath());
            sdkProperties.load(new FileInputStream(file));
        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            ordererDomain = sdkProperties.getProperty("orderer.domname");
            for (Map.Entry<Object, Object> x : sdkProperties.entrySet()) {
                final String key = x.getKey() + "";
                final String val = x.getValue() + "";
                if (key.startsWith(INTEGRATIONTESTS_ORG)) {
                    Matcher match = orgPat.matcher(key);
                    if (match.matches() && match.groupCount() == 1) {
                        String orgName = match.group(1).trim();
                        String mspId = val;
                        OrgEntity org = new OrgEntity(orgName);
                        org.setMspId(mspId);
                        orgs.put(orgName, org);
                    }
                }
            }

            for (Map.Entry<String, OrgEntity> entry : orgs.entrySet()) {
                final String orgName = entry.getKey();
                final OrgEntity org = entry.getValue();
                String domainName = sdkProperties.getProperty(INTEGRATIONTESTS_ORG + orgName + ".domname");
                org.setDomainName(domainName);
                String peerNames = sdkProperties.getProperty(INTEGRATIONTESTS_ORG + orgName + ".peer_locations");
                String[] ps = peerNames.split("[ \t]*,[ \t]*");
                for (String peer : ps) {
                    String[] nl = peer.split("[ \t]*@[ \t]*");
                    String peerName = nl[0];
                    String peerLocation = grpcTLSify(nl[1]);
                    PeerEntity peerEntity = new PeerEntity(peerName);
                    peerEntity.setOrg(org);
                    peerEntity.setPeerLocation(peerLocation);
                    peerEntity.setPeerProperties(getPeerProperties(domainName, peerName));
                    peers.put(peerName, peerEntity);
                    org.addPeer(peerEntity);
                }

                String eventHubNames = sdkProperties.getProperty(INTEGRATIONTESTS_ORG + orgName + ".eventhub_locations");
                ps = eventHubNames.split("[ \t]*,[ \t]*");
                for (String peer : ps) {
                    String[] nl = peer.split("[ \t]*@[ \t]*");
                    String peerName = nl[0];
                    String eventHubLocation = grpcTLSify(nl[1]);
                    peers.get(peerName).setPeerEventHubLocation(eventHubLocation);
                }

                String ordererNames = sdkProperties.getProperty(INTEGRATIONTESTS_ORG + orgName + ".orderer_locations");
                ps = ordererNames.split("[ \t]*,[ \t]*");
                for (String peer : ps) {
                    String[] nl = peer.split("[ \t]*@[ \t]*");
                    String ordererName = nl[0];
                    String ordererLocation = grpcTLSify(nl[1]);
                    Properties ordererProperties = getOrdererProperties(ordererDomain, ordererName);
                    OrdererEntity orderer = new OrdererEntity(ordererName, ordererLocation, ordererProperties);
                    orderers.put(ordererName, orderer);
                }

                String caLocation = httpTLSify(sdkProperties.getProperty((INTEGRATIONTESTS_ORG + orgName + ".ca_location")));
                try {
                    org.setCaClient(HFCAClient.createNewInstance(caLocation, getCaProperties(orgName)));
                } catch (Exception e) {
                    e.printStackTrace();
                }

                Admin admin = new Admin(orgName);
                admin.setMspId(org.getMspId());
                try {
                    admin.setEnrollment(org.getCaClient().enroll("admin", "adminpw"));
                } catch (EnrollmentException e) {
                    e.printStackTrace();
                } catch (InvalidArgumentException e) {
                    e.printStackTrace();
                }
                org.setAdmin(admin);

                try {
                    PeerAdmin peerAdmin = getPeerAdmin(orgName, org.getMspId(),
                            findFileSk(Paths.get(getChannelPath(), "/crypto-config/peerOrganizations/",
                                    domainName, String.format("/users/Admin@%s/msp/keystore/", domainName)).toFile()),
                            Paths.get(getChannelPath(), "/crypto-config/peerOrganizations/", domainName,
                                    String.format("/users/Admin@%s/msp/signcerts/Admin@%s-cert.pem", domainName,
                                            domainName)).toFile());
                    org.setPeerAdmin(peerAdmin);
                } catch (IOException e) {
                    e.printStackTrace();
                } catch (NoSuchAlgorithmException e) {
                    e.printStackTrace();
                } catch (NoSuchProviderException e) {
                    e.printStackTrace();
                } catch (InvalidKeySpecException e) {
                    e.printStackTrace();
                }
            }

            log.info("PropertiesConfigService orgs:{}", orgs);
        }
    }

    @Override
    public Collection<OrgEntity> listBaseOrgInfo() {
        return new ArrayList<OrgEntity>(orgs.values());
    }

    @Override
    public Collection<OrdererEntity> listBaseOrdererInfo() {
        return new ArrayList<OrdererEntity>(orderers.values());
    }

    @Override
    public Collection<String> listChannelName() {
        List<String> list = new ArrayList<>();
        for (int i = 1; i <= 12; i++) {
            list.add("trace" + i);
        }
        return list;
    }

    @Override
    public Collection<PeerEntity> listEndorserOfChannelOfChaincode(String channelName, String chaincodeName) {
        return new ArrayList<PeerEntity>(peers.values());
    }

    @Override
    public Collection<PeerEntity> listPeerOfChannel(String channelName) {
        return new ArrayList<PeerEntity>(peers.values());
    }

    private static void defaultProperty(String key, String value) {
        String ret = System.getProperty(key);
        if (ret != null) {
            sdkProperties.put(key, ret);
        } else {
            String envKey = key.toUpperCase().replaceAll("\\.", "_");
            ret = System.getenv(envKey);
            if (null != ret) {
                sdkProperties.put(key, ret);
            } else {
                if (null == sdkProperties.getProperty(key) && value != null) {
                    sdkProperties.put(key, value);
                }
            }
        }
    }

    public String getChannelPath() {
        return baasProperties.getChannelPath();
    }

    public Properties getPeerProperties(String domainName, String  peerName) {
//        String domainName = getDomainName(peerName);
        File cert = new File(getChannelPath() + "/crypto-config/peerOrganizations/" + domainName + "/peers" +
                "/" + peerName + "/tls/server.crt");
        if (!cert.exists()) {
            throw new RuntimeException(String.format("Missing cert file for: %s. Could not find at location: %s", peerName,
                    cert.getAbsolutePath()));
        }
        Properties ret = new Properties();
        ret.setProperty("pemFile", cert.getAbsolutePath());
        //      ret.setProperty("trustServerCertificate", "true"); //testing environment only NOT FOR PRODUCTION!
        ret.setProperty("hostnameOverride", peerName);
        ret.setProperty("sslProvider", "openSSL");
        ret.setProperty("negotiationType", "TLS");
        ret.setProperty("ordererWaitTimeMilliSecs", "3000000");
        return ret;
    }

    public Properties getOrdererProperties(String domainName,String ordererName) {
//        String domainName = getDomainName(ordererName);
        File cert = new File(getChannelPath() + "/crypto-config/ordererOrganizations/" + domainName + "/orderers" +
                "/" + ordererName + "/tls/server.crt");
        if (!cert.exists()) {
            throw new RuntimeException(String.format("Missing cert file for: %s. Could not find at location: %s", ordererName,
                    cert.getAbsolutePath()));
        }
        Properties ret = new Properties();
        ret.setProperty("pemFile", cert.getAbsolutePath());
        //      ret.setProperty("trustServerCertificate", "true"); //testing environment only NOT FOR PRODUCTION!
        ret.setProperty("hostnameOverride", ordererName);
        ret.setProperty("sslProvider", "openSSL");
        ret.setProperty("negotiationType", "TLS");
        ret.setProperty("ordererWaitTimeMilliSecs", "3000000");
        ret.put("grpc.NettyChannelBuilderOption.keepAliveTime", new Object[]{5L, TimeUnit.MINUTES});
        ret.put("grpc.NettyChannelBuilderOption.keepAliveTimeout", new Object[]{8L, TimeUnit.SECONDS});
        ret.put("grpc.NettyChannelBuilderOption.keepAliveWithoutCalls", new Object[]{true});
        return ret;
    }

    public Properties getCaProperties(String orgName) {
        String domainName = orgs.get(orgName).getDomainName();
        String CERTPATH = getChannelPath() + "/crypto-config/peerOrganizations/DNAME/tlsca/tlsca.DNAME-cert.pem";
        String cert = CERTPATH.replaceAll("DNAME", domainName);
        File cf = new File(cert);
        if (!cf.exists() || !cf.isFile()) {
            throw new RuntimeException("TEST is missing cert file " + cf.getAbsolutePath());
        }
        Properties properties = new Properties();
        properties.setProperty("pemFile", cf.getAbsolutePath());

        properties.setProperty("allowAllHostNames", "true");
        return properties;
    }

    private String getDomainName(final String name) {
        int dot = name.indexOf(".");
        if (-1 == dot) {
            return null;
        } else {
            return name.substring(dot + 1);
        }
    }

    private String grpcTLSify(String location) {
        location = location.trim();
        Exception e = Utils.checkGrpcUrl(location);
        if (e != null) {
            throw new RuntimeException(String.format("Bad TEST parameters for grpc url %s", location), e);
        }
        return location.replaceFirst("^grpc://", "grpcs://");
    }

    private String httpTLSify(String location) {
        location = location.trim();
        return location.replaceFirst("^http://", "https://");
    }

    public PeerAdmin getPeerAdmin(String orgName, String mspId, File privateKeyFile, File certificateFile) throws IOException, NoSuchAlgorithmException, NoSuchProviderException, InvalidKeySpecException {
        PeerAdmin peerAdmin = null;
        peerAdmin = new PeerAdmin(orgName);
        String certificate = new String(IOUtils.toByteArray(new FileInputStream(certificateFile)), "UTF-8");
        PrivateKey privateKey = getPrivateKeyFromBytes(IOUtils.toByteArray(new FileInputStream(privateKeyFile)));
        peerAdmin.setEnrollment(new EnrollmentEntity(privateKey, certificate));
        peerAdmin.setMspId(mspId);
        return peerAdmin;
    }

    private PrivateKey getPrivateKeyFromBytes(byte[] data) throws IOException, NoSuchProviderException, NoSuchAlgorithmException, InvalidKeySpecException {
        final Reader pemReader = new StringReader(new String(data));
        final PrivateKeyInfo pemPair;
        try (PEMParser pemParser = new PEMParser(pemReader)) {
            pemPair = (PrivateKeyInfo) pemParser.readObject();
        }
        PrivateKey privateKey = new JcaPEMKeyConverter().setProvider(new BouncyCastleProvider()).getPrivateKey(pemPair);
        return privateKey;
    }

    public File findFileSk(File directory) {
        File[] matches = directory.listFiles((dir, name) -> name.endsWith("_sk"));
        if (null == matches) {
            throw new RuntimeException(format("Matches returned null does %s directory exist?", directory.getAbsoluteFile().getName()));
        }
        if (matches.length != 1) {
            throw new RuntimeException(format("Expected in %s only 1 sk file but found %d", directory.getAbsoluteFile().getName(), matches.length));
        }
        return matches[0];
    }

}
